# -*- coding: utf-8 -*-

import re
import time
import logging
import datetime

from suds.client import Client
from suds.xsd.doctor import Import, ImportDoctor
from suds.transport.http import HttpAuthenticated

import poste
from poste import tol, xol

URL_TOL_TEST = 'http://sptest.posteitaliane.it/TelegrammaExtranet/WsTOL.svc'
URL_ROL_TEST = 'http://sptest.posteitaliane.it/ROLGC/RolService.svc'
URL_LOL_TEST = 'http://sptest.posteitaliane.it/LOLGC/LolService.svc'

URL_TOL_PROD = 'https://cewebservices.posteitaliane.it/TelegrammaExtranet/WsTOL.svc'
URL_ROL_PROD = 'https://cewebservices.posteitaliane.it/ROLGC/RolService.svc'
URL_LOL_PROD = 'https://cewebservices.posteitaliane.it/LOLGC/LolService.svc'

class TOLClient(object):
    
    #: an instance of suds Client
    schema = 'http://www.w3.org/2001/XMLSchema'
    SERVICE_NAME = 'TOL'

    filtri = ["http://ComunicazioniElettroniche.TOL.WS",]
    
    def __init__(self, username, password, service_level='PROD'):
        """Istanzia un client di connessione al servizio delle Poste italiane

        :param username: lo username da utilizzare
        :param password: la password da utilizzare
        :param service_level: TEST/PROD. 
        """
        imp = Import(self.schema)
        for filtro in self.filtri:
            imp.filter.add(filtro)
        self.URL = globals()['URL_%s_%s' % (self.SERVICE_NAME, service_level)] 

        doctor = ImportDoctor(imp)
        self.suds_client = Client(
            self.URL + '?WSDL' , doctor=doctor, username=username, password=password)
        self.suds_client.set_options(location=self.URL)

    def get_new_guid(self):
        "Fatti dare un id di richiesta. "
        ## NOTA: procedura molto differente da quella del ROL
        ## Ritorna direttamente un id
        return self.suds_client.service.GetIdRequest()

    def valida_destinatario(self, destinatario, id_request=None):
        assert isinstance(destinatario, tol.Destinatario), "%s must be poste.Destinatario" % destinatario
        if not id_request:
            id_request = self.get_new_guid()
        result = self.suds_client.service.RecipientValidation(
            recipient=destinatario.get_instance(self.suds_client), idRequest=id_request)
        if result.Result.ResType == "I":
            return result
        else:
            raise poste.RemoteRecipientValidationError(str(result))
        
    
class LOLClient(TOLClient):
    """
    Client per invio di raccomandate
    """

    SERVICE_NAME = 'LOL'
    filtri = ['http://ComunicazioniElettroniche.LOL.WS',
              'http://ComunicazioniElettroniche.XOL'
              ]
    autoconfirm = True
        
    def submit(self, documento, destinatari, mittente, guid=None, cliente=None):
        """Prima fase dell'invio di un documento ad un array di destinatari. Ogni
        operazione successiva relativa a questo ROL richiede che passi un
        tempo suff -intorno ai 60 secondi- che viene implementato nei metodi
        :meth:`valorizza`, :meth:`annulla`, :meth:`get_documento`,
        :meth:`confirm`, :meth:`preconferma`

        :param documento: oggetto :class:`poste.Documento` o filename
        :param destinatari: :class:`poste.Destinatario`: singolo, lista o set
        :param mittente: un :class:`poste.Mittente`
        :param guid: opzionale, l'id della richiesta
        :param cliente: identificativo del cliente

        :return: lol_submit creato, con ``lol_submit.response`` la response
                 ricevuta dal server

        """
        assert cliente, "Non so cosa sia 'cliente', ma mettilo..."

        if not guid:
            guid = self.get_new_guid()
            
        lol_submit = poste.LOLSubmit(documento, destinatari, mittente)
        lol_submit_instance = lol_submit.get_instance(self.suds_client)
        
        result = self.suds_client.service.Invio(
            IDRichiesta=guid, Cliente=cliente, LOLSubmit=lol_submit_instance)
        lol_submit.result = result

        if result.CEResult._Type == "I":
            return lol_submit

        raise poste.PosteSoapClientError(result)

    def get_new_guid(self):
        """
        Il client richiede un identificativo che rappresenta la transazione che vuole iniziare.

        :returns: IDRichiesta
        """
        result = self.suds_client.service.RecuperaIdRichiesta()
        if result.CEResult._Type == "I":
            return result.IDRichiesta
        else:
            raise poste.PosteSoapClientError(result)
    
    def get_receipt_id(self, XOLResponse):
        """Return a dict mapping id_destinatari (set by us) with 
        receipt_id (IdRicevuta) that is the recipient identification of the 
        web service. Not used internally but needed to track invoice status.

        :param XOLResponse: a response from a xol submit
        """
        # Usato in jmb.webposte!!
        dati = XOLResponse.ValidaDestinatariResults.DatiDestinatari
        res = {}
        for dest in dati.ValidaDestinatariResult:
            for d in dest.Destinatari.Destinatario:
                res[d._IdDestinatario] = d._IdRicevuta
        return res
        
    def valorizza(self, xol_submit=None, guid=None, verbose=False, direct=None, single=False):
        """Richiedi info relative al xol_submit - tempo minimo 35 secondi!! dall'invio
        La documentazione ufficiale chiede di aspettare 60 secondi circa

        Questo result e il ServizioEnquiryResponse sono definiti a pag 45
        result ha 2 attributi:

        + ``CEResult``
        + ``ServizioEnquiryResponse`` lista di 1 elemento [0] che poi ha:

         *   Richiesta (IDRichiesta e GuidUtente originali
         *   StatoLavorazione (dove peschiamo lo stato che determina questo ciclo)
         *   TipoVisualizzazione (InCorso / Archivio)
         *   NumeroPagine (del documento submitto)
         *   Totale
         *   OggettoValorizzazione (dettaglio della valorizzazione)
         *   CEResult

        :param xol_submit: :class:`poste.ROLSubmit`(:class:`poste.LOLSubmit`)
        :param guid: an id_richiesta (alias guid) relative to a submitted rol
        """ 
        richiesta = poste.Richiesta(guid=guid,
                                    xol_submit=xol_submit).get_instance(self.suds_client)
        # A pag 28, descrivendo la funzione invio, suggerisce di attendere 60 secondi per
        # recuperare i dati. Gli stati citati sono:
        # R: prezzato
        # N: Stato finale con errore
        # Di fatto la richiesta emessa subito riceve un W (risorsa non disponibile)
        status = ' '
        final_states =  'RNJSAVUWQ'
        while status not in final_states:  # pag 46, stati finali
            result = self.suds_client.service.Valorizza(Richieste=[richiesta,])
            # Vedi docstring per spiegazioni
            # ServizioEnquiryResponse, CEResult = result  NON FUNZIONA
            servEnq0 = result.ServizioEnquiryResponse[0]
            status = servEnq0.StatoLavorazione._Id
            if single:
                break
            if status not in final_states:
                print 'Riprovo Valorizza causa stato ' + status
                time.sleep(30)

        if not direct:
            time.sleep(35)  # Vedi nota sopra
            direct = False
        if status == 'N' and not single:
            self.parse_error(result)
 #           raise poste.PosteSoapClientError(result)


        return result
        
    def parse_error(self, result):
        "parse the error for simple cases"
#        import ipdb; ipdb.set_trace()

        try:
            result.ServizioEnquiryResponse[0].StatoLavorazione._Id == 'N'
            raise poste.RemoteValidationError(result)
        except AttributeError:
            pass
        
    def annulla(self, xol_submit=None, guid=None, verbose=False, direct=None):
        """Anulla un invio
        La documentazione ufficiale chiede di aspettare 60 secondi circa

        :param xol_submit: :class:`poste.ROLSubmit`
        :param guid: a GuidUser to retrieve information on a submitted rol 
        """ 
        richiesta = poste.Richiesta(
            guid=guid, xol_submit=xol_submit).get_instance(self.suds_client)

        result = self.suds_client.service.Annulla(richiesta)
        
        if not result.CEResult._Type == 'I':
            raise PosteSoapClientError(result)
        return result

    def get_documento(self, xol_submit=None, guid=None, 
                     document_save_path=None, verbose=False, direct=None):
        """Richiedi il documento submitto. RecuperaDCS nella documentazione ufficiale.
        Leggi le note in :meth:`valorizza` per capire quando sono disponibili i dati

        :param xol_submit: :class:`poste.ROLSubmit`
        :param guid: a GuidUser to retrieve information on a submitted rol 
        :param document_save_path: path dove salvare il documento prelevato (opzionale)
        """ 
        richiesta = poste.Richiesta(guid=guid, xol_submit=xol_submit).get_instance(self.suds_client)
        # Leggi nota in valorizza per capire il motivo del ritardo
        status = ' '
        
        while status not in 'I':
            result = self.suds_client.service.RecuperaDCS(Richiesta=richiesta)
            status = result.CEResult._Type 
            if not direct:
                time.sleep(35)  # Vedi nota sopra
                direct = False

        if document_save_path:
            decode_file(risposta.Documento.Immagine, output_path=document_save_path)

        return result

    def preconferma(self, xol_submit=None, guid=None, autoconfirm=None):
        """
        PreConferma un invio. Necessario per confermare.

        :param xol_submit: :class:`poste.ROLSubmit` (:class:`poste.LOLSubmit`)
        :param guid: a GuidUser to retrieve information on a submitted rol/lol
        :param autoconfirm: boolean: se True, non serve eseguire conferma successiva
        """
        self.valorizza(xol_submit, guid)
        preconfermareq = poste.Richiesta(
            guid=guid, xol_submit=xol_submit).get_instance(self.suds_client)
        # autoconfirm: prevale riga di comando poi self.cutoconfirm
        autoconfirm = self.autoconfirm if autoconfirm == None else autoconfirm
        result = self.suds_client.service.PreConferma(
            Richieste=preconfermareq, autoConferma=autoconfirm)
        if not result.CEResult._Type == 'I':
            raise poste.PosteSoapClientError(result)
        return result

    def confirm(self, result=None, xol_submit=None, guid=None, autoconfirm=None):
        """
        Conferma un invio già pre-confermato

        :param result: il valore di ritorno della preconferma, parametro suff.
        :param xol_submit: :class:`poste.ROLSubmit` (:class:`poste.LOLSubmit`),
            richiede automaticamnete una preconferma
        :param guid: a GuidUser to retrieve information on a submitted rol 
        :param autoconfirm: boolean: se True, non serve eseguire conferma successiva
        """
        autoconfirm = self.autoconfirm if autoconfirm == None else autoconfirm
        if not result:
            result = self.preconferma(xol_submit, guid, autoconfirm=autoconfirm)
        if autoconfirm:
            return result
        op = result.Valorizzazione.OpzionePagamento[0]
        creation_name = '{http://ComunicazioniElettroniche.XOL}DatiTransazione'
        dt = self.suds_client.factory.create(creation_name)
        result = self.suds_client.service.Conferma(IdOrdine=result.IdOrdine,
                                         OpzionePagamento=op,
                                         DatiTransazione=dt)
        if not result.CEResult._Type == 'I':
            raise poste.PosteSoapClientError(result)
        return result

    def get_stato_ordine(self, guid=None, xol_submit=None):
        """
        Recupera lo stato di un ordine
        """
        if not guid:
            guid = xol_submit.guid
        assert guid, "Guid e' necessario per recuperare lo stato"
    
        richiesta = poste.Richiesta(guid=guid).get_instance(self.suds_client)

        result = self.suds_client.service.RecuperaStatoOrdine(IdOrdine=richiesta)
        return result
        
    def get_stato_richiesta(self, guid=None, xol_submit=None):
        """
        Recupera lo stato di una richiesta
        """
        if not guid:
            guid = xol_submit.guid
        assert guid, "GUID e' necessario per recuperare lo stato"
    
        richiesta = poste.Richiesta(guid=guid).get_instance(self.suds_client)

        result = self.suds_client.service.RecuperaStatoIdRichiesta(IdRichiesta=richiesta)
        return result

    def valida_destinatario(self, destinatario, guid=None, test=False):
        return self.valida_destinatari([destinatario], guid, test)

    def valida_destinatari(self, destinatari, guid=None, test=False):
        """
        Valida molti destinatari

        :param destinatari: lista di destinatari
        :param guid: facoltativo (id_richiesta)
        :param test: se True, ritorna anche l'id_richiesta 
        """
        if not guid:
            guid = self.get_new_guid()
            
        aod = self.suds_client.factory.create('ArrayOfDestinatario')
        for dest in destinatari:
            aod.Destinatario.append(dest.get_instance(self.suds_client))            

        result = self.suds_client.service.ValidaDestinatari(
            IDRichiesta=guid, Destinatari=aod)

        if result.CEResult._Type == "I":
            if test:
                return result, guid
            else:
                return result
        else:
            raise poste.RemoteRecipientValidationError(
                result, guid)
    
class ROLClient(LOLClient):
    """
    Client per l'invio di lettere
    """
    SERVICE_NAME = 'ROL'
    filtri = ['http://ComunicazioniElettroniche.ROL.WS',
              'http://ComunicazioniElettroniche.XOL'
              ]

    def submit(self, documento, destinatari, mittente, guid=None, cliente='',
              ricevuta_ritorno=None):
        """Prima fase dell'invio di un documento ad un array di destinatari. Ogni
        operazione successiva relativa a questo ROL richiede che passi un
        tempo suff -circa 60 secondi- che viene implementato nei metodi
        :meth:`valorizza`, :meth:`annulla`, :meth:`get_documento`,
        :meth:`confirm`, :meth:`preconferma`

        :param documento: oggetto :class:`poste.Documento` o filename
        :param destinatari: :class:`poste.Destinatario`: singolo, lista o set
        :param mittente: un :class:`poste.Mittente`
        :param guid: opzionale, l'id della richiesta
        :param cliente: identificativo del cliente
        :param ricevuta_ritorno: Mittente (lista) per ricevuta di ritorno
        
        :return: :class:`poste.ROLSubmit` creato, con ``rol_submit.response`` la response
                 ricevuta dal server

        """
        if not guid:
            guid = self.get_new_guid()
            
        rol_submit = poste.ROLSubmit(documento, destinatari, mittente,
                                     ricevuta_ritorno=ricevuta_ritorno)
        rol_submit_instance = rol_submit.get_instance(self.suds_client)
        
        result = self.suds_client.service.Invio(
            IDRichiesta=guid, Cliente=cliente, ROLSubmit=rol_submit_instance)
        rol_submit.result = result
        rol_submit.sent_datetime = datetime.datetime.now()

        if result.CEResult._Type == "I":
            return rol_submit

        raise poste.PosteSoapClientError(result)


    def get_stato_invii_per_id(self, lista_id):
        """
        Recupera lo stato di un invio. id non corretti vengno semplicemente ignorati.

        :param lista_id: lista di IdRichiesta (lista di stringhe)
        :return: Un dizionario le cui chiavi sono gli id (solo quelli validi, ovvero
                 corrispondenti a invii confermati) ed i valori sono quanto ritornato

        Vedi :meth:`get_stato_invii` per una descrizione del valore di ritorno
        
        """
        assert isinstance(lista_id, (tuple, list, set)), "%s deve essere una lista di id" % lista_id

        lista_richieste = []
        for guid in lista_id:
            lista_richieste += [poste.Richiesta(
                guid=guid).get_instance(self.suds_client)]

        result = self.suds_client.service.StatoInviiPerID(Richieste=lista_richieste)
        return self._wrap_stato_invii(result)

    def _wrap_stato_invii(self, result):
        "return result in a nicer way: DataEsito as datetime"
        res = {}

        if hasattr(result, 'ArrayDiRichieste'):
            info_request = {}
            for item in result.ArrayDiRichieste:
                for info in item.StatoRichieste:
                    info.DataEsito = datetime.datetime.strptime(info.DataEsito, '%d/%m/%Y %H:%M:%S')
                    info_request[info.NumeroServizio] = info
                res[item.IdRichiesta] = info_request
        if hasattr(result, 'StatoRichieste'):
            for item in result.StatoRichieste:
                item.DataEsito = datetime.datetime.strptime(item.DataEsito, '%d/%m/%Y %H:%M:%S')
                res[item.NumeroServizio] = item
        return res

    def get_stato_invii(self,):
        """
        Recupera lo stato di un invio. Gli id non corretti vengono ignorati.

        :return: Un dizionario le cui chiavi sono gli id (solo quelli validi, ovvero
        corrispondenti a invii confermati) ed il valore un dizionario con chiave il numero 
        di raccomandata e  quanto ritornato::

            {'fcf1f462-028d-46c3-aa44-36cb0267cfd6': {
                '500000003510' :   (StatoRichiesta){
                                      StatoType = "00"
                                      StatoDescrizione = "accettata online"
                                      NumeroServizio = "500000003510"
                                      DataEsito = 2012-02-28 17:46:36
                }
            }

        La DataEsito è un datetime::

          res['fd06df5a-df73-4d84-ad06-1eadf2b9234d'].DataEsito = datetime(...)

        Il Numero di servizio è il numero della raccomandata.
        Gli stati possibili sono descritti nel cap 4.2.1.23::

        :00: Accettata online (transitorio)
        :01: Consegnato       (definitivo)
        :02: In giacenza      (transitorio)
        :03: In restituzione  (definitivo)
        :99: In lavorazione   (transitorio)
        
        """
        result = self.suds_client.service.StatoInvii()
        return self._wrap_stato_invii(result)

    def get_stato_invii_per_numero(self, id_list):
        """
        Recupera lo stato di un invio dal numero di raccomandata
        
        :param id_list: list of certified mail number (Numeri raccomandata)

        :return: a dict whose keys are the id (numero raccomandata) of id_list and the value is 
                 a StatoRichiesta.

        Vedi :meth:`get_stato_invii` per una descrizione più approfondita del valore di ritorno
        
        """

        result = self.suds_client.service.StatoInviiPerNumero(NumeroMissive=id_list)
        return self._wrap_stato_invii(result)

    def valida_destinatario(self, destinatario, guid=None, test=False):
        """
        Submit una richiesta di validazione indirizzi

        :param destinatari: una lista di class:`poste.base.Destinatario`
        :param test: boolean: if running for tests, guid is returned as well.
               Required by PosteOnlineTest
        """
        return self.valida_destinatari([destinatario,], guid=None, test=test)

